# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/00_core.ipynb (unless otherwise specified).

__all__ = ['make_differentiation_matrices', 'est_wrapped_gradient', 'p_shrink', 'make_laplace_kernel', 'unwrap']

# Cell
import numpy as np
from scipy import sparse as sp
from scipy.fft import dctn, idctn
from .loading import load_interferogram

# Cell
def make_differentiation_matrices(
    rows, columns, boundary_conditions="neumann", dtype=np.float32
):
    """Generate derivative operators as sparse matrices.

    Matrix-vector multiplication is the fastest way to compute derivatives
    of large arrays, particularly for images. This function generates
    the matrices for computing derivatives. If derivatives of the same
    size array will be computed more than once, then it generally is
    faster to compute these arrays once, and then reuse them.

    The three supported boundary conditions are 'neumann' (boundary
    derivative values are zero), 'periodic' (the image ends wrap around
    to beginning), and 'dirichlet' (out-of-bounds elements are zero).
    'neumann' seems to work best for solving the unwrapping problem.

    Source:
    https://github.com/rickchartrand/regularized_differentiation/blob/master/regularized_differentiation/differentiation.py
    """
    bc_opts = ["neumann", "periodic", "dirichlet"]
    bc = boundary_conditions.strip().lower()
    if bc not in bc_opts:
        raise ValueError(f"boundary_conditions must be in {bc_opts}")

    # construct derivative with respect to x (axis=1)
    D = sp.diags([-1.0, 1.0], [0, 1], shape=(columns, columns), dtype=dtype).tolil()

    if boundary_conditions.lower() == bc_opts[0]:  # neumann
        D[-1, -1] = 0.0
    elif boundary_conditions.lower() == bc_opts[1]:  # periodic
        D[-1, 0] = 1.0
    else:
        pass

    S = sp.eye(rows, dtype=dtype)
    Dx = sp.kron(S, D, "csr")

    # construct derivative with respect to y (axis=0)
    D = sp.diags([-1.0, 1.0], [0, 1], shape=(rows, rows), dtype=dtype).tolil()

    if boundary_conditions.lower() == bc_opts[0]:
        D[-1, -1] = 0.0
    elif boundary_conditions.lower() == bc_opts[1]:
        D[-1, 0] = 1.0
    else:
        pass

    S = sp.eye(columns, dtype=dtype)
    Dy = sp.kron(D, S, "csr")

    return Dx, Dy

# Cell
def est_wrapped_gradient(
    arr, Dx=None, Dy=None, boundary_conditions="neumann", dtype=np.float32
):
    """Estimate the wrapped gradient of `arr` using differential operators `Dx, Dy`
    Adjusts the grad. to be in range [-pi, pi]
    """
    rows, columns = arr.shape
    if Dx is None or Dy is None:
        Dx, Dy = make_differentiation_matrices(
            rows, columns, boundary_conditions=boundary_conditions, dtype=dtype
        )

    phi_x = (Dx @ arr.ravel()).reshape((rows, columns))
    phi_y = (Dy @ arr.ravel()).reshape((rows, columns))
    # Make wrapped adjustmend (eq. (2), (3))
    idxs = np.abs(phi_x) > np.pi
    phi_x[idxs] -= 2 * np.pi * np.sign(phi_x[idxs])
    idxs = np.abs(phi_y) > np.pi
    phi_y[idxs] -= 2 * np.pi * np.sign(phi_y[idxs])
    return phi_x, phi_y

# Cell
def p_shrink(X, lmbda=1, p=0, epsilon=0):
    """p-shrinkage in 1-D, with mollification."""

    mag = np.sqrt(np.sum(X ** 2, axis=0))
    nonzero = mag.copy()
    nonzero[mag == 0.0] = 1.0
    mag = (
        np.maximum(
            mag
            - lmbda ** (2.0 - p) * (nonzero ** 2 + epsilon) ** (p / 2.0 - 0.5),  # noqa
            0,
        )
        / nonzero
    )

    return mag * X

# Cell
def make_laplace_kernel(rows, columns, dtype='float32'):
    """Generate eigenvalues of diagonalized Laplacian operator

    Used for quickly solving the linear system ||D \Phi - phi|| = 0

    References:
    Numerical recipes, Section 20.4.1, Eq. 20.4.22 is the Neumann case
    or https://elonen.iki.fi/code/misc-notes/neumann-cosine/
    """
    # Note that sign is reversed from numerical recipes eq., since
    # here since our operator discretization sign reversed
    xi_y = (2 - 2 * np.cos(np.pi * np.arange(rows) / rows)).reshape((-1, 1))
    xi_x = (2 - 2 * np.cos(np.pi * np.arange(columns) / columns)).reshape((1, -1))
    eigvals = xi_y + xi_x

    with np.errstate(divide="ignore"):
        K = np.nan_to_num(1 / eigvals, posinf=0, neginf=0)
    return K.astype(dtype)

# Cell
def unwrap(
    f_wrapped,
    phi_x=None,
    phi_y=None,
    max_iters=500,
    tol=np.pi / 5,
    lmbda=1,
    p=0,
    c=1.3,
    dtype="float32",
    debug=False,
    #     boundary_conditions="neumann",
):
    """Unwrap interferogram phase

    Parameters
    ----------
        f_wrapped (ndarray): wrapped phase image (interferogram)
        phi_x (ndarray): estimate of the x-derivative of the wrapped phase
            If not passed, will compute using `est_wrapped_gradient`
        phi_y (ndarray): estimate of the y-derivative of the wrapped phase
            If not passed, will compute using `est_wrapped_gradient`
        max_iters (int): maximum number of ADMM iterations to run
        tol (float): maximum allowed change for any pixel between ADMM iterations
        lmbda (float): splitting parameter of ADMM. Smaller = more stable, Larger = faster convergence.
        p (float): value used in shrinkage operator
        c (float): acceleration constant using in updating lagrange multipliers in ADMM
        dtype: numpy datatype for output
        debug (bool): print diagnostic ADMM information
    """
    rows, columns = f_wrapped.shape
    num = rows * columns

    if dtype is None:
        dtype = f_wrapped.dtype
    else:
        f_wrapped = f_wrapped.astype(dtype)

    boundary_conditions = "neumann"
    if debug:
        print(f"Making Dx, Dy with BCs={boundary_conditions}")
    Dx, Dy = make_differentiation_matrices(
        *f_wrapped.shape, boundary_conditions=boundary_conditions
    )

    if phi_x is None or phi_y is None:
        phi_x, phi_y = est_wrapped_gradient(f_wrapped, Dx, Dy, dtype=dtype)

    # Lagrange multiplier variables
    Lambda_x = np.zeros_like(phi_x, dtype=dtype)
    Lambda_y = np.zeros_like(phi_y, dtype=dtype)

    # aux. variables for ADMM, holding difference between
    # unwrapped phase gradient and measured gradient from igram
    w_x = np.zeros_like(phi_x, dtype=dtype)
    w_y = np.zeros_like(phi_y, dtype=dtype)

    F_old = np.zeros_like(f_wrapped)

    # Get K ready once for solving linear system
    K = make_laplace_kernel(rows, columns, dtype=dtype)

    for iteration in range(max_iters):

        # update Unwrapped Phase F: solve linear eqn in fourier domain
        # rhs = dx.T @ phi[0].ravel() + dy.T @ phi[1].ravel()
        rx = w_x.ravel() + phi_x.ravel() - Lambda_x.ravel()
        ry = w_y.ravel() + phi_y.ravel() - Lambda_y.ravel()
        RHS = Dx.T * rx + Dy.T * ry
        # Use DCT for neumann:
        rho_hat = dctn(RHS.reshape(rows, columns), type=2, norm='ortho', workers=-1)
        F = idctn(rho_hat * K, type=2, norm='ortho', workers=-1)

        # calculate x, y gradients of new unwrapped phase estimate
        Fx = (Dx @ F.ravel()).reshape(rows, columns)
        Fy = (Dy @ F.ravel()).reshape(rows, columns)

        input_x = Fx - phi_x + Lambda_x
        input_y = Fy - phi_y + Lambda_y
        w_x, w_y = p_shrink(
            np.stack((input_x, input_y), axis=0), lmbda=lmbda, p=p, epsilon=0
        )

        # update lagrange multipliers
        Lambda_x += c * (Fx - phi_x - w_x)
        Lambda_y += c * (Fy - phi_y - w_y)

        change = np.max(np.abs(F - F_old))
        if debug:
            print(f"Iteration:{iteration} change={change}")

        if change < tol or np.isnan(change):
            break
        else:
            F_old = F

    if debug:
        print(f"Finished after {iteration} with change={change}")
    return F


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/01_loading.ipynb (unless otherwise specified).

__all__ = ['load_dem_rsc', 'load_interferogram', 'load_gdal', 'RSC_KEY_TYPES']

# Cell
import os
import collections
import numpy as np
try:
    import gdal
    _GDAL_INSTALLED = True
except ImportError:
    print("WARNING: gdal not installed, can only load binary '.int' files")
    print("Use 'conda install -c conda-forge gdal' to install for other formats.")
    _GDAL_INSTALLED = False

# Cell

RSC_KEY_TYPES = [
    ("width", int),
    ("file_length", int),
    ("x_first", float),
    ("y_first", float),
    ("x_step", float),
    ("y_step", float),
    ("x_unit", str),
    ("y_unit", str),
    ("z_offset", int),
    ("z_scale", int),
    ("projection", str),
]


def load_dem_rsc(filename, lower=False, **kwargs):
    """Loads and parses the .dem.rsc file
    Args:
        filename (str) path to either the .dem or .dem.rsc file.
            Function will add .rsc to path if passed .dem file
        lower (bool): make keys of the dict lowercase
    Returns:
        dict: dem.rsc file parsed out, keys are all caps
    example file:
    WIDTH         10801
    FILE_LENGTH   7201
    X_FIRST       -157.0
    Y_FIRST       21.0
    X_STEP        0.000277777777
    Y_STEP        -0.000277777777
    X_UNIT        degrees
    Y_UNIT        degrees
    Z_OFFSET      0
    Z_SCALE       1
    PROJECTION    LL
    """

    # Use OrderedDict so that upsample_dem_rsc creates with same ordering as old
    output_data = collections.OrderedDict()
    # Second part in tuple is used to cast string to correct type

    rsc_filename = (
        "{}.rsc".format(filename) if not filename.endswith(".rsc") else filename
    )
    with open(rsc_filename, "r") as f:
        for line in f.readlines():
            for field, num_type in RSC_KEY_TYPES:
                if line.startswith(field.upper()):
                    output_data[field] = num_type(line.split()[1])

    if lower:
        output_data = {k.lower(): d for k, d in output_data.items()}
    return output_data


def load_interferogram(filename, dtype=np.complex64, columns=None, rsc_file=None, band=1):
    """Load binary complex interferogram file.

    Can either pass the number of columns, or a .rsc file with metadata
    """
    if not filename.endswith(".int"):
        return load_gdal(filename, band=band)

    data = np.fromfile(filename, dtype)
    if columns is None:
        if rsc_file is None:
            try:
                rsc_file = filename + ".rsc"
                assert os.path.exists(rsc_file)
                rsc_data = load_dem_rsc(rsc_file, lower=True)
                columns = rsc_data["width"]
            except AssertionError:
                raise ValueError(
                    f"columns or rsc_file not passed, {filename + '.rsc'} does not exist."
                )

    return data.reshape((-1, columns))


def load_gdal(filename, band=1):
    """If not loading a binary .int file, try using GDAL"""
    if not _GDAL_INSTALLED:
        raise ValueError(
            "gdal needed to load filetypes other than '.int'."
            " Use 'conda install -c conda-forge gdal' to install for other formats."
        )
    ds = gdal.Open(filename)
    band = ds.GetRasterBand(band)
    image = band.ReadAsArray()
    band, ds = None, None
    return image


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/02_cli.ipynb (unless otherwise specified).

__all__ = ['get_cli_args', 'main']

# Cell
import argparse
import os
import numpy as np
from .core import unwrap
from .loading import load_interferogram

# Cell


def get_cli_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", help="Name of interferogram file to load")
    parser.add_argument(
        "--outname", "-o", help="Name of output file to save (default = `filename`.unw)"
    )
    parser.add_argument(
        "--max-iters",
        default=500,
        type=int,
        help="maximum number of ADMM iterations to run (default = %(default)s)",
    )
    parser.add_argument(
        "--tol",
        default=np.pi / 10,
        type=float,
        help="maximum allowed change for any pixel between ADMM iterations (default = %(default)s)",
    )
    parser.add_argument(
        "--lmbda",
        default=1,
        type=float,
        help="splitting parameter of ADMM. Smaller = more stable, Larger = faster "
        "convergence. (default = %(default)s)",
    )
    parser.add_argument(
        "--p",
        default=0,
        type=float,
        help="value used in shrinkage operator (default = %(default)s)",
    )
    parser.add_argument(
        "--c",
        default=1.3,
        type=float,
        help="acceleration constant using in updating lagrange multipliers in ADMM "
        "(default = %(default)s)",
    )
    parser.add_argument(
        "--dtype",
        default="float32",
        help="numpy datatype of filename (default = %(default)s)",
    )
    parser.add_argument(
        "--debug", action="store_true", help="print diagnostic ADMM information"
    )
    return parser.parse_args()


def main():
    arg_dict = vars(get_cli_args())
    inname = arg_dict.pop("filename")
    igram = load_interferogram(inname)
    mag = np.abs(igram)
    phase = np.angle(igram)
    outname = arg_dict.pop("outname")
    if outname is None:
        base, ext = os.path.splitext(inname)
        outname = base + ".unw"

    if not outname.endswith(".unw"):
        raise NotImplementedError(
            "Only saving as binary .unw is implemented currently."
        )
    # TODO: save as other types?
    unw_phase = unwrap(phase, **arg_dict)
    unw_with_mag = np.hstack((mag, unw_phase))
    unw_with_mag.tofile(outname)


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"make_differentiation_matrices": "00_core.ipynb",
         "est_wrapped_gradient": "00_core.ipynb",
         "p_shrink": "00_core.ipynb",
         "make_laplace_kernel": "00_core.ipynb",
         "unwrap": "00_core.ipynb",
         "load_dem_rsc": "01_loading.ipynb",
         "load_interferogram": "01_loading.ipynb",
         "load_gdal": "01_loading.ipynb",
         "RSC_KEY_TYPES": "01_loading.ipynb",
         "get_cli_args": "02_cli.ipynb",
         "main": "02_cli.ipynb"}

modules = ["core.py",
           "loading.py",
           "cli.py"]

doc_url = "https://scottstanie.github.io/spurs/"

git_url = "https://github.com/scottstanie/spurs/tree/master/"

def custom_doc_links(name):
    return None
